@@ -29,6 +29,13 @@
 
 import struct
 import numpy as np
+
+try:
+    from numpy.compat import asbytes, asstr
+except:
+    asbytes = str
+    asstr = str
+
 import tempfile
 import zlib
 import warnings
@@ -44,6 +51,8 @@ DTYPE_DICT[6] = '>c8'
 DTYPE_DICT[7] = '|O'
 DTYPE_DICT[8] = '|O'
 DTYPE_DICT[9] = '>c16'
+DTYPE_DICT[10] = '|O'
+DTYPE_DICT[11] = '|O'
 DTYPE_DICT[12] = '>u2'
 DTYPE_DICT[13] = '>u4'
 DTYPE_DICT[14] = '>i8'
@@ -73,7 +82,7 @@ def _align_32(f):
     '''Align to the next 32-bit position in a file'''
 
     pos = f.tell()
-    if pos % 4 <> 0:
+    if pos % 4 != 0:
         f.seek(pos + 4 - pos % 4)
     return
 
@@ -91,7 +100,7 @@ def _read_bytes(f, n):
 
 def _read_byte(f):
     '''Read a single byte'''
-    return np.uint8(struct.unpack('>B', f.read(4)[0])[0])
+    return np.uint8(struct.unpack('>B', f.read(4)[:1])[0])
 
 
 def _read_long(f):
@@ -147,15 +156,22 @@ class Pointer(object):
         return
 
 
+class ObjectPointer(Pointer):
+    '''Class used to define object pointers'''
+    pass
+
+
 def _read_string(f):
     '''Read a string'''
     length = _read_long(f)
     if length > 0:
         chars = _read_bytes(f, length)
         _align_32(f)
+        chars = asstr(chars)
     else:
-        chars = None
-    return np.str(chars)
+        warnings.warn("warning: empty strings are now set to '' instead of None")
+        chars = ''
+    return chars
 
 
 def _read_string_data(f):
@@ -163,18 +179,19 @@ def _read_string_data(f):
     length = _read_long(f)
     if length > 0:
         length = _read_long(f)
-        string = _read_bytes(f, length)
+        string_data = _read_bytes(f, length)
         _align_32(f)
     else:
-        string = None
-    return np.str(string)
+        warnings.warn("warning: empty strings are now set to '' instead of None")
+        string_data = ''
+    return string_data
 
 
 def _read_data(f, dtype):
     '''Read a variable with a specified data type'''
     if dtype==1:
-        if _read_int32(f) <> 1:
-            raise Exception("Error occured while reading byte variable")
+        if _read_int32(f) != 1:
+            raise Exception("Error occurred while reading byte variable")
         return _read_byte(f)
     elif dtype==2:
         return _read_int16(f)
@@ -199,7 +216,7 @@ def _read_data(f, dtype):
     elif dtype==10:
         return Pointer(_read_int32(f))
     elif dtype==11:
-        raise Exception("Object reference type not implemented")
+        return ObjectPointer(_read_int32(f))
     elif dtype==12:
         return _read_uint16(f)
     elif dtype==13:
@@ -249,6 +266,13 @@ def _read_structure(f, array_desc, struct_desc):
             else:
                 structure[col['name']][i] = _read_data(f, dtype)
 
+    # Reshape structure if needed
+    if array_desc['ndims'] > 1:
+        warnings.warn("warning: multi-dimensional structures are now correctly reshaped")
+        dims = array_desc['dims'][:int(array_desc['ndims'])]
+        dims.reverse()
+        structure = structure.reshape(dims)
+
     return structure
 
 
@@ -262,8 +286,8 @@ def _read_array(f, typecode, array_desc):
 
         if typecode == 1:
             nbytes = _read_int32(f)
-            if nbytes <> array_desc['nbytes']:
-                raise Exception("Error occured while reading byte array")
+            if nbytes != array_desc['nbytes']:
+                raise Exception("Error occurred while reading byte array")
 
         # Read bytes as numpy array
         array = np.fromstring(f.read(array_desc['nbytes']), \
@@ -289,7 +313,7 @@ def _read_array(f, typecode, array_desc):
 
     # Reshape array if needed
     if array_desc['ndims'] > 1:
-        dims = array_desc['dims'][:array_desc['ndims']]
+        dims = array_desc['dims'][:int(array_desc['ndims'])]
         dims.reverse()
         array = array.reshape(dims)
 
@@ -328,7 +352,7 @@ def _read_record(f):
         rectypedesc = _read_typedesc(f)
 
         varstart = _read_long(f)
-        if varstart <> 7:
+        if varstart != 7:
             raise Exception("VARSTART is not 7")
 
         if rectypedesc['structure']:
@@ -465,7 +489,7 @@ def _read_arraydesc(f):
         arraydesc['dims'] = []
         for d in range(arraydesc['nmax']):
             v = _read_long(f)
-            if v <> 0:
+            if v != 0:
                 raise Exception("Expected a zero in ARRAY_DESC")
             arraydesc['dims'].append(_read_long(f))
 
@@ -482,15 +506,19 @@ def _read_structdesc(f):
     structdesc = {}
 
     structstart = _read_long(f)
-    if structstart <> 9:
+    if structstart != 9:
         raise Exception("STRUCTSTART should be 9")
 
     structdesc['name'] = _read_string(f)
-    structdesc['predef'] = _read_long(f)
+    predef = _read_long(f)
     structdesc['ntags'] = _read_long(f)
     structdesc['nbytes'] = _read_long(f)
 
-    if structdesc['predef'] & 1 == 0:
+    structdesc['predef'] = predef & 1
+    structdesc['inherits'] = predef & 2
+    structdesc['is_super'] = predef & 4
+
+    if not structdesc['predef']:
 
         structdesc['tagtable'] = []
         for t in range(structdesc['ntags']):
@@ -509,18 +537,24 @@ def _read_structdesc(f):
             if tag['structure']:
                 structdesc['structtable'][tag['name']] = _read_structdesc(f)
 
-        STRUCT_DICT[structdesc['name']] = (structdesc['tagtable'], \
-                                           structdesc['arrtable'], \
-                                           structdesc['structtable'])
+        if structdesc['inherits'] or structdesc['is_super']:
+            structdesc['classname'] = _read_string(f)
+            structdesc['nsupclasses'] = _read_long(f)
+            structdesc['supclassnames'] = []
+            for s in range(structdesc['nsupclasses']):
+                structdesc['supclassnames'].append(_read_string(f))
+            structdesc['supclasstable'] = []
+            for s in range(structdesc['nsupclasses']):
+                structdesc['supclasstable'].append(_read_structdesc(f))
+
+        STRUCT_DICT[structdesc['name']] = structdesc
 
     else:
 
         if not structdesc['name'] in STRUCT_DICT:
             raise Exception("PREDEF=1 but can't find definition")
 
-        structdesc['tagtable'], \
-        structdesc['arrtable'], \
-        structdesc['structtable'] = STRUCT_DICT[structdesc['name']]
+        structdesc = STRUCT_DICT[structdesc['name']]
 
     return structdesc
 
@@ -546,6 +580,67 @@ def _read_tagdesc(f):
     return tagdesc
 
 
+def _replace_heap(variable, heap):
+
+    if isinstance(variable, Pointer):
+
+        while isinstance(variable, Pointer):
+
+            if variable.index == 0:
+                variable = None
+            else:
+                variable = heap[variable.index]
+
+        replace, new = _replace_heap(variable, heap)
+
+        if replace:
+            variable = new
+
+        return True, variable
+
+    elif isinstance(variable, np.core.records.recarray):
+
+        # Loop over records
+        for ir, record in enumerate(variable):
+
+            replace, new = _replace_heap(record, heap)
+
+            if replace:
+                variable[ir] = new
+
+        return False, variable
+
+    elif isinstance(variable, np.core.records.record):
+
+        # Loop over values
+        for iv, value in enumerate(variable):
+
+            replace, new = _replace_heap(value, heap)
+
+            if replace:
+                variable[iv] = new
+
+        return False, variable
+
+    elif isinstance(variable, np.ndarray):
+
+        # Loop over values if type is np.object_
+        if variable.dtype.type is np.object_:
+
+            for iv in range(variable.size):
+
+                replace, new = _replace_heap(variable.item(iv), heap)
+
+                if replace:
+                    variable.itemset(iv, new)
+
+        return False, variable
+
+    else:
+
+        return False, variable
+
+
 class AttrDict(dict):
     '''
     A case-insensitive dictionary with access via item, attribute, and call
@@ -623,27 +718,27 @@ def read(file_name, idict=None, python_dict=False,
         variables = AttrDict()
 
     # Open the IDL file
-    f = file(file_name, 'rb')
+    f = open(file_name, 'rb')
 
     # Read the signature, which should be 'SR'
     signature = _read_bytes(f, 2)
-    if signature <> 'SR':
+    if signature != asbytes('SR'):
         raise Exception("Invalid SIGNATURE: %s" % signature)
 
     # Next, the record format, which is '\x00\x04' for normal .sav
     # files, and '\x00\x06' for compressed .sav files.
     recfmt = _read_bytes(f, 2)
 
-    if recfmt == '\x00\x04':
+    if recfmt == asbytes('\x00\x04'):
         pass
 
-    elif recfmt == '\x00\x06':
+    elif recfmt == asbytes('\x00\x06'):
 
         if verbose:
             print "IDL Save file is compressed"
 
         if uncompressed_file_name:
-            fout = file(uncompressed_file_name, 'w+b')
+            fout = open(uncompressed_file_name, 'w+b')
         else:
             fout = tempfile.NamedTemporaryFile(suffix='.sav')
 
@@ -651,14 +746,14 @@ def read(file_name, idict=None, python_dict=False,
             print " -> expanding to %s" % fout.name
 
         # Write header
-        fout.write('SR\x00\x04')
+        fout.write(asbytes('SR\x00\x04'))
 
         # Cycle through records
         while True:
 
             # Read record type
             rectype = _read_long(f)
-            fout.write(struct.pack('>l', rectype))
+            fout.write(struct.pack('>l', int(rectype)))
 
             # Read position of next record and return as int
             nextrec = _read_uint32(f)
@@ -678,16 +773,16 @@ def read(file_name, idict=None, python_dict=False,
             pos = f.tell()
 
             # Decompress record
-            string = zlib.decompress(f.read(nextrec-pos))
+            rec_string = zlib.decompress(f.read(nextrec-pos))
 
             # Find new position of next record
-            nextrec = fout.tell() + len(string) + 12
+            nextrec = fout.tell() + len(rec_string) + 12
 
             # Write out record
             fout.write(struct.pack('>I', int(nextrec % 2**32)))
             fout.write(struct.pack('>I', int((nextrec - (nextrec % 2**32)) / 2**32)))
             fout.write(unknown)
-            fout.write(string)
+            fout.write(rec_string)
 
         # Close the original compressed file
         f.close()
@@ -719,8 +814,9 @@ def read(file_name, idict=None, python_dict=False,
     # Find all variables
     for r in records:
         if r['rectype'] == "VARIABLE":
-            while isinstance(r['data'], Pointer):
-                r['data'] = heap[r['data'].index]
+            replace, new = _replace_heap(r['data'], heap)
+            if replace:
+                r['data'] = new
             variables[r['varname'].lower()] = r['data']
 
     if verbose:
@@ -761,7 +857,7 @@ def read(file_name, idict=None, python_dict=False,
         rectypes = [r['rectype'] for r in records]
 
         for rt in set(rectypes):
-            if rt <> 'END_MARKER':
+            if rt != 'END_MARKER':
                 print " - %i are of type %s" % (rectypes.count(rt), rt)
         print "-"*50
 
